#ifndef _MESH_HPP_
#define _MESH_HPP_

#include <fstream.h>	//only necessary for mesh text in-/output in MESH_IO

//turing machine code starts here. --------------------------------by jph 2004

//#define MESH_BAR		//include barycentric coordinate computations
//#define MESH_GL			//include openGL display
//#define MESH_GST		//include openGL display and phantom
//#define MESH_IO			//include mesh text file input and output
//#define MESH_DEBUG	//define to receive certain warnings

#define MESH_VERT      0			//indexes the ind[.] array which stores vertex indices for each shape
#define MESH_EDGE      1
#define MESH_TRIA      2
#define MESH_QUAD      3
#define MESH_TETR      4
#define MESH_OCTA      5
#define MESH_PRIS      6
#define MESH_CUBE      7
#define MESH_MAX       8

#define MESH_VERT_     1			//the same again for bitwise occupation, as used in drawing routine
#define MESH_EDGE_     2                                             
#define MESH_TRIA_     4
#define MESH_QUAD_     8
#define MESH_TETR_     16
#define MESH_OCTA_     32
#define MESH_PRIS_     64
#define MESH_CUBE_     128
#define MESH_ALL       255
static int MESH_OCCUP  []={1,2,4,8,16,32,64,128};	//bitwise access in for loops

#define MESH_ZIPPER    1			//for example subdivide(MESH_ZIPPER|MESH_LINEAR|MESH_AVERAGE)
#define MESH_LINEAR    2			//not all combinations are possible, I forgot which,
#define MESH_AVERAGE   4			//but subdivide(MESH_LINEAR) works.
#define MESH_NORMAL    7

#define MESH_FORCE     256		//attribute for PHANToM operation
#define MESH_JAYA      1024		//file input from jaya mesh
#define MESH_TXT       2048		//file input from text format mesh

#define MESH_NORMALS   256		//has currently no effect
#define MESH_VERTICES  512
#define MESH_TOPOLOGY  1024
#define MESH_NOCOLOR   1024
#define MESH_NODEPTH   2048		//disables depth buffer in mesh drawing routine
#define MESH_OPAQUE    8192		//displays the shapes not transparent
#define MESH_LABEL     16384	//labels vertices with their index
#define MESH_EDGES     32768	//uses also next 2 bits to encode size
#define MESH_GIANT     536870912 //alternatively 67108864

static int    MESH_D       []={0,3,12,240,MESH_ALL};	//used for addressing all of dim i
static int    MESH_COMP    []={1,2,3,4,4,6,6,8};			//number of vertices at shape
static int    MESH_ADD_VPF []={0,0,1,2,2,3,4,7};			//number of vertices essencially added by face REVISE!!!
static int    MESH_REPRO   []={1,2,4,4,4,6,8,8};			//number of face type after lin sub
static int    MESH_REPOSIT []={1,3,6,9,10,19,18,27};	//vertices in face at lin sub
static int    MESH_DIMS    []={0,1,2,2,3,3,3,3};			//dimensionality of shapes
static int    MESH_JAYAIND []={9,MESH_VERT,MESH_EDGE,MESH_TRIA,MESH_TETR,9,MESH_OCTA,9}; //the ones with nine dont matter
static int    MESH_CLASS   []={MESH_VERT,MESH_EDGE,MESH_TRIA,MESH_QUAD,MESH_TETR,MESH_TETR,MESH_PRIS,MESH_CUBE}; //vertex classes
static double MESH_AVERAGEC[]={        1,       .5,1.0/6.0,     .25,1.0/14.0, 1.0/14.0,1.0/12.0,1.0/8.0};

#ifdef MESH_GL
static int MESH_GLTYP   []={GL_POINTS,GL_LINES,GL_TRIANGLES,GL_QUADS}; //vertex classes
#endif

//#define MESH_FASTAVG		//if defined not C^2 but a little faster and good results
//	proc=[
//		0 2 2 2 3 5
//		1 3 4 5 5 4];

static int MESH_ZIPIND  []={MESH_TRIA,MESH_QUAD,MESH_TETR,MESH_PRIS,MESH_PRIS,MESH_CUBE,MESH_OCTA,MESH_EDGE,MESH_TRIA,MESH_QUAD}; //boundary comes from
static int MESH_ZIPCLASS[]={        2,        3,        4,        5,        6,        7,        4,        9,        9,        9}; //boundary global class 9 - irrelevant
static int MESH_ZIPSIZE []={        2,        2,        3,        3,        4,        4,        3,        2,        3,        4};	//boundary length: edge edge tri tri quad ...
#define MESH_ZIPLEN 6		//this is six, not seven because octa is merged with the tetr, should be =mesh_proc.size[1]

matrix<int>* mesh_rep[4*MESH_MAX];	//edges quads
array<float> mesh_colr;
matrix<float> mesh_shp[MESH_MAX];
matrix<int> mesh_proc,mesh_uni[MESH_MAX],mesh_bas[MESH_MAX],mesh_zip[10],mesh_swp[MESH_MAX];
matrix<int> mesh_bar[MESH_MAX],mesh_brn[MESH_MAX],mesh_nrm[MESH_MAX+2],mesh_mea[MESH_MAX];	//last two store tris of PRIS and CUBE (a little bit of waste)

void mesh_init(void) {
	int c1,c2,sz[2];
	char fnc[80];
	mesh_proc=matrix<int> ("tab/mesh_proc.tab");
	mesh_colr=array<float>("tab/mesh_colr.tab");
	for (c1=0;c1<MESH_MAX;c1++) {
		sprintf(fnc,"tab/mesh_shp%i.tab",c1); mesh_shp[c1]=matrix<float>(fnc);
		sprintf(fnc,"tab/mesh_edg%i.tab",c1); mesh_rep[4*c1+0]=new matrix<int>(fnc); //edges
		sprintf(fnc,"tab/mesh_qua%i.tab",c1); mesh_rep[4*c1+1]=new matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_hex%i.tab",c1); mesh_rep[4*c1+2]=new matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_cub%i.tab",c1); mesh_rep[4*c1+3]=new matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_uni%i.tab",c1); mesh_uni[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_bas%i.tab",c1); mesh_bas[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_zip%i.tab",c1); mesh_zip[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_bar%i.tab",c1); mesh_bar[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_brn%i.tab",c1); mesh_brn[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_nrm%i.tab",c1); mesh_nrm[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_mea%i.tab",c1); mesh_mea[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_swp%i.tab",c1); mesh_swp[c1]=matrix<int>(fnc);
	}
	for (c1=0;c1<4*MESH_MAX;c1++)
		if (!mesh_rep[c1]->data||mesh_rep[c1]->numel==0) {
			delete mesh_rep[c1];
			mesh_rep[c1]=NULL;
		}
	for (c1=8;c1<10;c1++) {
		sprintf(fnc,"tab/mesh_zip%i.tab",c1); mesh_zip[c1]=matrix<int>(fnc);
		sprintf(fnc,"tab/mesh_nrm%i.tab",c1); mesh_nrm[c1]=matrix<int>(fnc);
	}
};

int  mesh_hash_numel;	//yes, a global variable, that is modified from within
											//member functions of the mesh class so that our function
											//pointer trick for the hash function works

unsigned int mesh_hash(array<int> mat) {			//hash function #1
	int c1,cnt=0;													
	for (c1=0;c1<mat.numel;c1++) cnt+=mat[c1];
	return cnt%mesh_hash_numel;
};
unsigned int mesh_trivial(array<int> mat) {		//hash function #2
	return mat.data[0];
};

//--------------------------------------------------------------------------------------
// MESH CLASS, globally the first, that does what it does, by jph
// <type> is type of vertex coordinates (usually float or double)
// main functionality is volumetric subdivision
//--------------------------------------------------------------------------------------
template <class type> class mesh {
	public:
		mesh<type>(void);
		mesh<type>(int sv);												//sv is dimensionality of each vertex
		mesh<type>(char *fn);											//retrieves mesh from binary files fn.extentions
		mesh<type>(const mesh<type>& msh);
		~mesh<type>(void);												//deletes mesh - of course
		void setUp(void);													//upon instanciation, retrieve matrices depending on type
		void store(char *fn);											//saves mesh in binary mode to fn.extentions
		int  addVerts(array<type> v); 						//appends vertices to current vertex array returns current last vertex index
		void addFaces(int typ,matrix<int> f);			//appends faces, dito		
		void addBoundary(void);										//adds boundary after certain rules
		void subdivide(int opt);									//main function (choose opt from ZIPPER,LINEAR,AVERAGE,NORMAL)		
		void apply(matrix<type> mat);							//applies the 4x4 affine transformation mat to the xyz part (that either you (with mode) or it determines!) of the mesh		
		void empty(int opt);											//empties mesh (shape&|vert)		e.g. opt=MESH_EDGE_|MESH_TRIA_|MESH_VERTICES
		void clean(int opt);											//opt=MESH_VERTICES|MESH_TOPOLOGY, removes degenerates shapes and/or unreferenced vertices (hard remove)
		void content(char *str);									//overwrites str to contain number of each shape type and vertices (good as a compact info)				
		const mesh<type>& operator= (const mesh<type>& a);	//equal
		array<int> *ind[MESH_MAX];								//topology
		matrix<type> *arr;												//vertex array		
		int xyz_ofs;															//optional to define, initial guess upon initialization

#ifdef MESH_BAR
#include "mesh_bar.hpp"
#endif
		
#ifdef MESH_GL
#include "mesh_gl.hpp"
#endif

#ifdef MESH_GST
#include "mesh_gst.hpp"
#endif

#ifdef MESH_IO
#include "mesh_io.hpp"
#endif

	private:
		void arr_zip(void);												//DATA STRUCT OPERATION (NOT subdiv!)
		void zip_arr(void);												//DATA STRUCT OPERATION (NOT subdiv!)
		void sub_zip(void);												//does the zipping
		void sub_lin(void);												//needs zip hash -> vertices remain in zip hash
		void sub_avg(void);												//needs zip hash -> vertices collapse to arr (zip is deleted afterwards)
		bool match(array<int> &in1,array<int> &in2,array<int> &ord);	//super important for zipping: orients (or rejects) a boundary pair, such that matrix mult is possible
		int numVertEst(void);											//guesses the number of inserted vertices by sub_lin, important for hash table
		int getAdr(array<int> *badge);						//I don't know anymore
		void mean(array<int> &mat,array<int> *badge);	//...either
		array<int> boundary(array<int> &mat);			//just a boundary key (inverts order abcd->adcb)		
		hash<array<int>,matrix<type> > *zip; 			//big vertex hash		
		matrix<type> *mesh_avg[MESH_MAX],*mesh_mul[MESH_ZIPLEN];
		array<type> *mesh_cor[MESH_MAX];					//correction factor		
		hash<array<int>,int> *adr;								//indices to vertex hash
		int count,status;													//count=arr->size[1]; and gets increased over sub_lin while vertices are added (hard)
};

//--------------------------------------------------------------------------------------
template <class type> mesh<type>::mesh<type>(void) {
	int c1,sz[2]; sz[1]=0;
	for (c1=0;c1<MESH_MAX;c1++) {
		sz[0]=MESH_COMP[c1];
		ind[c1]=new array<int>(2,sz,0);
	}
	arr=NULL;
	setUp();	
}
template <class type> mesh<type>::mesh<type>(int sv) {
	int c1,sz[2]; sz[1]=0;
	for (c1=0;c1<MESH_MAX;c1++) {
		sz[0]=MESH_COMP[c1];
		ind[c1]=new array<int>(2,sz,0);
	}
	arr=new matrix<type>(sv,0);
	setUp();	
}

template <class type> mesh<type>::mesh<type>(char *fn) {
	int c1,sz[2]; sz[1]=0;
	char fnc[80];
	for (c1=0;c1<MESH_MAX;c1++) {
		sprintf(fnc,"%s_%i.ind",fn,c1);	ind[c1]=new array<int>(fnc);		
		if (!ind[c1]->data) {
			delete ind[c1];
			sz[0]=MESH_COMP[c1];
			ind[c1]=new array<int>(2,sz,0);
		}
	}
	sprintf(fnc,"%s.arr",fn); arr=new matrix<type>(fnc);	
	if (!arr->data) {
		delete arr;
		arr=NULL;
	}
	setUp();
};

template <class type> mesh<type>::mesh(const mesh<type>& msh) {
	int c1;
	for (c1=0;c1<MESH_MAX;c1++) ind[c1]=new array<int>(*msh.ind[c1]);
	arr=new matrix<type>(*msh.arr);
	xyz_ofs=msh.xyz_ofs;
#ifdef MESH_GL
	mode=msh.mode;
#endif
	setUp();
};

template <class type> mesh<type>::~mesh<type>(void) {
	int c1;	
	for (c1=0;c1<MESH_MAX;c1++) {		
		delete ind[c1];				
		delete mesh_avg[c1];
		delete mesh_cor[c1];
	}
	for (c1=0;c1<MESH_ZIPLEN;c1++) delete mesh_mul[c1];	
	if (arr) delete arr;	
#ifdef MESH_BAR
	if (ocp) delete ocp;
	if (bar) delete bar;
	if (spa) delete spa;
#endif
}

template <class type> void mesh<type>::setUp(void) {
	int c1; char fnc[80];
	//int sz=mesh_proc.size[1];
	array_output=false;
	for (c1=0;c1<MESH_MAX;c1++) {
		sprintf(fnc,"tab/mesh_avg%i%i.tab",sizeof(type),c1); mesh_avg[c1]=new matrix<type>(fnc);
		sprintf(fnc,"tab/mesh_cor%i%i.tab",sizeof(type),c1); mesh_cor[c1]=new array<type>(fnc);		
	}
	for (c1=0;c1<MESH_ZIPLEN;c1++) {
		sprintf(fnc,"tab/mesh_mul%i%i.tab",sizeof(type),c1); mesh_mul[c1]=new matrix<type>(fnc);
	}
	array_output=true;
	xyz_ofs=0;
	adr=NULL;
	zip=NULL;
	count=0; status=0;
#ifdef MESH_GL
	tex=0;
	mode=GL_V3F;
	if (arr->size[0]== 6) {xyz_ofs=3; mode=GL_C3F_V3F;			}
	if (arr->size[0]== 8) {xyz_ofs=5; mode=GL_T2F_N3F_V3F;	}
	if (arr->size[0]==10) {xyz_ofs=7; mode=GL_C4F_N3F_V3F;	}
#endif
#ifdef MESH_BAR
	atr_ofs=0;
	ocp=NULL;
	bar=NULL;
	spa=NULL;
#endif
}
//--------------------------------------------------------------------------------------
template <class type> void mesh<type>::store(char *fn) {
	int c1;	char fnc[80];
	for (c1=0;c1<MESH_MAX;c1++) 
		if (ind[c1]->numel) {
			sprintf(fnc,"%s_%i.ind",fn,c1); ind[c1]->store(fnc);
		}
	sprintf(fnc,"%s.arr",fn); arr->store(fnc);
}

template <class type> int mesh<type>::addVerts(array<type> v) {
	if (!arr) arr=new matrix<type>(v.size[0],0);
	v.stretch(arr->size[0],0);
	*arr|=v;
#ifdef MESH_BAR
	int c1;
	if (atr_ofs)
		for (c1=arr->size[1]-v.size[1];c1<arr->size[1];c1++)
			memset(&arr->data[atr_ofs+c1*arr->size[0]],0,4);
#endif
	return arr->size[1]-v.size[1];
}

template <class type> void mesh<type>::addFaces(int typ,matrix<int> f) {
	*ind[typ]|=f;
}

template <class type> void mesh<type>::addBoundary(void) {//the first part of this code is almost identical to sub_zip, maybe merge into sep func
	int c1,c2,c3,c4,c5,mc1,ic1,p1,p2,sz[2],sv=arr->size[0];	
	hash<array<int>,int> bnd(arr->size[1],mesh_trivial);	//hosts boundary faces around vertex key is [min rest representant#], content is unrelevant	
	for (c1=0;c1<MESH_ZIPLEN+1;c1++) {										//extract boundary for all in MESH_ZIPIND
		ic1=MESH_ZIPIND [c1];
		c2 =MESH_ZIPSIZE[c1]+1;									
		array<int> bne(1,&MESH_ZIPSIZE[c1],0);	//holds temporarily to rotate and then adds to
		array<int> key(1,             &c2 ,0);	//... this to be entered in hash table bnd
		for (c2=0;c2<ind[ic1]->numel;c2+=ind[ic1]->size[0])	{				//for all faces			
			for (c3=0;c3<mesh_zip[c1].numel;c3+=MESH_ZIPSIZE[c1]) {	//for all boundary
				for (c4=0;c4<MESH_ZIPSIZE[c1];c4++)
					bne[c4]=ind[ic1]->data[c2+mesh_zip[c1].data[c3+c4]];
				bne.min(mc1);												//detects minimum
				bne.lshift(mc1);										//brings it to front (keeps orientation)
				memcpy(key.data,bne.data,MESH_ZIPSIZE[c1]*4);				
				key[MESH_ZIPSIZE[c1]]=MESH_ZIPSIZE[c1];	//HERE DIFFERENT ARRAY !!!				append zip class of boundary piece				
				if (!bnd.remove(boundary(key))) bnd.store(key,c4);	//other way around is faster:-) c4 is only void carrier
			}
		}
	}																					//<-- here bnd contains usually edges, tria & quad	
	c1=3; array<int> tri(1,&c1,0);
	c1=4; array<int> bne(1,&c1,0);
				array<int> key(1,&c1,0); key(3)=3;	//aha!?
	int occ[4];
	bucket<array<int>,int> buc;								//this time we dig into PUBLIC hash structure
	for (c1=0;c1<arr->size[1];c1++)						//this loops counts how many of each type will be added in total
		while (bnd.table->data[c1]->token(buc))
			if (buc.key.numel==5) {								//is quad?
				memcpy(bne^0,buc.key^0,16);
				for (c2=0;c2<4;c2++) {
					bne.lshift(1);
					memcpy(tri^0,bne^0,12);
					tri.min(mc1);											//detects minimum
					tri.lshift(mc1);									//brings it to front (keeps orientation)
					memcpy(key^0,tri^0,12);					
					if (!bnd.remove(boundary(key))) occ[c2]=0; else occ[c2]=1; //removes triangles touching quad
				}
				if (occ[0]+occ[1]+occ[2]+occ[3]) {
					bnd.table->data[c1]->remove();		//removes quad, in the following just the right triangles are added (a shorter version of the code is possible)
					if (occ[0]&&!occ[2]) {
						bne.lshift(3); memcpy(key^0,bne^0,12); bnd.store(key,c3);	}
					if (occ[2]&&!occ[0]) {
						bne.lshift(1);
						memcpy(key^0,bne^0,12); bnd.store(key,c3); }
					if (occ[1]&&!occ[3]) {
						memcpy(key^0,bne^0,12); bnd.store(key,c3); }
					if (occ[3]&&!occ[1]) {
						bne.lshift(2);
						memcpy(key^0,bne^0,12); bnd.store(key,c3); }
				}
			}																			//<-- from here bnd is final and transformed to hard shapes:
	matrix<int> num(MESH_TETR,1);							//quad is biggest boundary that can happen, therefore it will happen	
	for (c1=0;c1<arr->size[1];c1++)						//this loops counts how many of each type will be added in total
		while (bnd.table->data[c1]->token(buc))	//final boundary peace
			num.data[buc.key.numel-2]++;	
	matrix<int> shp[MESH_TETR];
	for (c1=0;c1<MESH_TETR;c1++)							//for vert/edge/tria/quad
		shp[c1]=matrix<int>(MESH_COMP[c1],num.data[c1]);
	memset(num^0,0,16);
	for (c1=0;c1<arr->size[1];c1++)
		while (bnd.table->data[c1]->token(buc)) {
			c2=buc.key.numel-2;
			memcpy(shp[c2]^num[c2],buc.key.data,4*(c2+1));
			num[c2]+=c2+1;
		}
	for (c1=0;c1<MESH_TETR;c1++) addFaces(c1,shp[c1]);//here we finally add :-)		
}

template <class type> void mesh<type>::subdivide(int opt) {
	int sv=arr->size[0];
	count=arr->size[1];																//offset from which vertices are added in sub_lin
	zip=new hash<array<int>,matrix<type> >(numVertEst(),mesh_trivial);
	arr_zip();																				//flushes face-[type/no] at each vertex	
	//printf("zip");
	if (opt&MESH_ZIPPER) sub_zip();										//refines the zip hash		
	//printf(" lin");
	if (opt&MESH_LINEAR) sub_lin();										//performs the topology split	
	delete arr; arr=new matrix<type>(sv,count);
	//printf(" avg");
	if (opt&MESH_AVERAGE) sub_avg(); else zip_arr();	//matrix multiplies shapes and correction term for surface	
	//printf(" del");
	delete zip; zip=NULL;	
	//printf(" done\n");
}
//--------------------------------------------------------------------------------------
template <class type> void mesh<type>::apply(matrix<type> mat) {
	int c1,sv=arr->size[0],len=3*sizeof(type);
	matrix<type> tmp(4,arr->size[1],"ones");
	for (c1=0;c1<arr->size[1];c1++)
		memcpy(&tmp.data[4*c1],&arr->data[xyz_ofs+c1*sv],len);	
	matrix<type> res=mat*tmp;
	for (c1=0;c1<arr->size[1];c1++)
		memcpy(&arr->data[xyz_ofs+c1*sv],&res.data[4*c1],len);
}

template <class type> void mesh<type>::empty(int opt) {
	int c1,sz[2],sv=arr->size[0];
	for (c1=0;c1<MESH_MAX;c1++)
		if (ind[c1]->numel&&opt&MESH_OCCUP[c1]) {
			delete ind[c1];
			sz[0]=MESH_COMP[c1]; sz[1]=0;
			ind[c1]=new array<int>(2,sz,0);
		}
	if (opt&MESH_VERTICES) {
		delete arr;
		arr=new matrix<type>(sv,0);
	}
}

template <class type> void mesh<type>::clean(int opt) {
	int c1,c2,cnt,sv=arr->size[0];	
	if (opt&MESH_TOPOLOGY) {
		for (c1=0;c1<MESH_MAX;c1++) {
			array<int> shp(1,&MESH_COMP[c1],0);
			array<int> occ(1,&ind[c1]->size[1],0);
			cnt=0;
			for (c2=0;c2<ind[c1]->size[1];c2++) {
				shp.numel=MESH_COMP[c1];
				memcpy(shp.data,&ind[c1]->data[cnt],MESH_COMP[c1]*4);
				shp.unique();
				occ.data[c2]=shp.numel==MESH_COMP[c1];
				cnt+=MESH_COMP[c1];
			}
			array<int> *idt=new array<int>(ind[c1]->queue2(occ.find()));
			delete ind[c1];
			ind[c1]=idt;
		}
	}
	if (opt&MESH_VERTICES) {
		cnt=1;
		array<int> occ(1,&arr->size[1],0);
		for (c1=0;c1<MESH_MAX;c1++)
			for (c2=0;c2<ind[c1]->numel;c2++) occ[ind[c1]->data[c2]]=1;
		matrix<type> *art=new matrix<type>(arr->queue2(occ.find()));
		for (c1=0;c1<arr->size[1];c1++) if (occ[c1]) occ[c1]=cnt++;	
		for (c1=0;c1<MESH_MAX;c1++)
			for (c2=0;c2<ind[c1]->numel;c2++) ind[c1]->data[c2]=occ[ind[c1]->data[c2]]-1;
		delete arr;
		arr=art;
	}
}

template <class type> void mesh<type>::content(char *str) {
	sprintf(str,"[%i,%i|2D:%i,%i|3D:%i,%i,%i,%i]::%ix%i:xyz@%i",
		ind[MESH_VERT]->size[1],
		ind[MESH_EDGE]->size[1],
		ind[MESH_TRIA]->size[1],
		ind[MESH_QUAD]->size[1],
		ind[MESH_TETR]->size[1],
		ind[MESH_OCTA]->size[1],
		ind[MESH_PRIS]->size[1],
		ind[MESH_CUBE]->size[1],
		arr->size[0],
		arr->size[1],
		xyz_ofs);
}
template <class type> const mesh<type>& mesh<type>::operator= (const mesh<type>& msh) {
	int c1;
	if (arr==msh.arr) return *this;	
	for (c1=0;c1<MESH_MAX;c1++) {
		delete ind[c1];
		delete mesh_avg[c1];
		delete mesh_cor[c1];
	}
	for (c1=0;c1<MESH_ZIPLEN;c1++) delete mesh_mul[c1];
	if (arr) delete arr;
#ifdef MESH_BAR
	if (ocp) delete ocp;
	if (bar) delete bar;
#endif	
	for (c1=0;c1<MESH_MAX;c1++) ind[c1]=new array<int>(*msh.ind[c1]);
	arr=new matrix<type>(*msh.arr);
	xyz_ofs=msh.xyz_ofs;
#ifdef MESH_GL
	mode=msh.mode;
#endif
	setUp();
	return *this;
};

//--------------------------------------------------------------------------------------
template <class type> void mesh<type>::arr_zip(void) {
	int c1=2,c2,sv=arr->size[0],svT=arr->size[0]*sizeof(type);
	array<int> typ(1,&c1,0);						//zip [vertex number==address in hash & type]
	matrix<type> tmp(arr->size[0],1);		//temporary vertex
	matrix<type> vrt(arr->size[0],1);		//temporary vertex
	for (c1=0;c1<MESH_MAX;c1++) {
		typ.data[1]=MESH_CLASS[c1];
		for (c2=0;c2<ind[c1]->numel;c2++) {
			typ.data[0]=ind[c1]->data[c2];
			memcpy(vrt.data,&arr->data[sv*ind[c1]->data[c2]],svT);
			zip->store(typ,vrt);
		}
	}
}

template <class type> void mesh<type>::zip_arr(void) {
	int c1,cnt=0;
	bucket<array<int>,matrix<type> > b;	
	for (c1=0;c1<count;c1++) {
		if (zip->table->data[c1]->token(b))
			memcpy(&arr->data[cnt],b.data.data,sizeof(type)*arr->size[0]);
		cnt+=arr->size[0];
	}
}

template <class type> void mesh<type>::sub_zip(void) {
	int c1,c2,c3,c4,c5,mc1,ic1,p1,p2,sz[2],sv=arr->size[0],svT=arr->size[0]*sizeof(type);	
	hash<array<int>,int> bnd(arr->size[1],mesh_trivial);	//hosts boundary faces around vertex key is [min rest representant#], content is unrelevant	
	for (c1=0;c1<MESH_ZIPLEN+1;c1++) {				//extract boundary from tria->cube->octa
		ic1=MESH_ZIPIND  [c1];									//true shape index
		c2 =MESH_ZIPSIZE [c1]+1;								//true boundary length +1
		array<int> bne(1,&MESH_ZIPSIZE[c1],0);	//holds temporarily to rotate and then adds to
		array<int> key(1,             &c2 ,0);	//... this to be entered in hash table bnd
		for (c2=0;c2<ind[ic1]->numel;c2+=ind[ic1]->size[0])				//for all faces
			for (c3=0;c3<mesh_zip[c1].numel;c3+=MESH_ZIPSIZE[c1]) {	//for all boundary
				for (c4=0;c4<MESH_ZIPSIZE[c1];c4++)
					bne[c4]=ind[ic1]->data[c2+mesh_zip[c1].data[c3+c4]];
				if (MESH_ZIPCLASS[c1]==MESH_PRIS) {	//this of course sad four performance (only for Qpris)
					bne.max(mc1); 										//get max and check pos of at
					if (mc1&1) bne.data[mc1]|=MESH_GIANT;	//mark if at==1|3
				}
				bne.min(mc1);												//detects minimum
				bne.lshift(mc1);										//brings it to front (keeps orientation)
				memcpy(key.data,bne.data,MESH_ZIPSIZE[c1]*4);
				key[MESH_ZIPSIZE[c1]]=MESH_ZIPCLASS[c1]-2;					//append zip class of boundary piece (why can this line not be outside again?)
				if (!bnd.remove(boundary(key))) bnd.store(key,c4);	//other way around is faster:-) c4 is only void carrier
			}
	}																					//<-- here bnd contains usually edge, tria & quad
	int mid,mps;
	bucket<array<int>,int> buc;
	array<int> qud(1,new int(5),0);						//this guy has the mission to carry a quad with 2nd smalles vert index in front
	for (c1=arr->size[1]-1;c1>=0;c1--) 				//loop over all verts from the back
		while (bnd.table->data[c1]->token(buc)) //duplicate remaining quads
			if (buc.key.numel==5)	{								//exploits that quads satisfy that due to [0123|x]
				mps=0;
				mid=arr->size[1];										//upper bound for vertex index
				for (c2=1;c2<4;c2++)
					if (buc.key[c2]<mid) {mid=buc.key[c2]; mps=c2; }
				memcpy( qud.data       ,&buc.key.data[mps],(4-mps)<<2);
				memcpy(&qud.data[4-mps], buc.key.data     ,    mps<<2);
				qud.data[4]=buc.key.data[4];
				bnd.store(qud,c2);									//add transformed duplicate
			}	
	//<-- here bnd contains edge tria & quadA & quadB
	matrix<int> num(MESH_ZIPLEN,1);						//counters for boundary types at one vertex
	hash<array<int>,matrix<type> > acc(arr->size[1],mesh_trivial); //holds new vertices
	c1=2; array<int> typ(1,&c1,0);						//acc [vertex number/pair ID (which in proc)]	// zip [vertex number/pairtype]
	matrix<type> vrt(sv+1,1),tmp(sv+1,1);			//vertex plus counter in last coordinate
	matrix<int> tqs[MESH_ZIPLEN];							//buffer to hold boundary
	for (c1=0;c1<MESH_ZIPLEN;c1++)						//for tria/quad/toc/tpris/qpris/cube
		tqs[c1]=matrix<int>(MESH_ZIPSIZE[c1],40);	//40 is HARDCODED LIMIT OF "VALENCE" !!! while 4/6 are ordinary "valences".
	array<int> ord;
	matrix<type> tab[3];											//collects vertices for matrixmult choice dep on ord.numel
	array<int> in1[3],in2[3];									//ferries for boundary match
	for (c1=2;c1<5;c1++) {
		tab[c1-2]=matrix<type>(sv,c1);
		in1[c1-2]=array<int>(1,&c1,0);
		in2[c1-2]=array<int>(1,&c1,0);
	}
	for (c1=0;c1<arr->size[1];c1++) {					//loop over all verts		
		memset(num.data,0,MESH_ZIPLEN<<2);			//prepare counters that track boundary appearance
		while (bnd.table->data[c1]->token(buc)) {	//flush all boundary at vertex into appropriate array
			ic1=buc.key.data[buc.key.numel-1];		//well whats that? Lets see... ??? Uh, yeah, bad trick: MESH_ZIPCLASS-2 awaits us
			memcpy(tqs[ic1]^num[ic1]*MESH_ZIPSIZE[ic1],buc.key.data,MESH_ZIPSIZE[ic1]<<2); //thank you, memory copy machine
			num[ic1]++;
		} 																			//<-- here the boundary is properly stored in buffer tqs (intuitive name)		
		for (c2=0;c2<mesh_proc.size[1];c2++) {	//distinguish cases according to procedure array
			p1=mesh_proc(0,c2);										//encode which two boundaries will be compared
			p2=mesh_proc(1,c2);										//...
			typ.data[1]=c2;												//inject PAIR ID in "accumulator hash" key
			if (p2==4) in2[2].size[1]=MESH_PRIS;	//if p2 is quad pris then make in2.size[1]=pris (hidden pris indicator for match)
						else in2[2].size[1]=0;			
			for (c3=0;c3<num[p1];c3++) {
				memcpy(in1[MESH_ZIPSIZE[p1]-2].data,tqs[p1]^c3*MESH_ZIPSIZE[p1],MESH_ZIPSIZE[p1]<<2);
				for (c4=0;c4<num[p2];c4++) {				//one might think this is some kind of O(m*n) algorithm
					memcpy(in2[MESH_ZIPSIZE[p2]-2].data,tqs[p2]^c4*MESH_ZIPSIZE[p2],MESH_ZIPSIZE[p2]<<2);
					if (match(in1[MESH_ZIPSIZE[p1]-2],in2[MESH_ZIPSIZE[p2]-2],ord)) {
						mc1=ord.numel-2;
						for (c5=0;c5<ord.numel;c5++)		//copy the vertices returned by match
							memcpy(tab[mc1]^c5*sv,*arr^sv*ord[c5],svT);
						tab[mc1]*=*mesh_mul[c2];				//computes vertices delta						
						for (c5=0;c5<ord.numel;c5++) {	//add to all vertices in acc
							memcpy(vrt.data,tab[mc1]^sv*c5,svT);	//vertex data							
							vrt.data[sv]=1;								//hidden counter in last coordinate
							typ.data[0]=ord.data[c5];			//the very vertex number
							if (acc.find(typ,tmp)) vrt+=tmp;	//part of THE delta(p1,p2)
							acc.prior(typ,vrt);
						}
					}
				}
			}
		}
	} //<-- here all deltas are accumulated //--> next: divide over valence and join if neccessary
	type f,val;
	bucket<array<int>,matrix<type> > get;				//this time we dig into PUBLIC hash structure
	for (c1=0;c1<arr->size[1];c1++) {						//for all verts
		typ.data[0]=c1;
		while (acc.table->data[c1]->token(get))	{	//get has [0:vertex address,1:pair ID,2:vertex type (YET! in 0..2)]			
			//printf("\n#%i proc:%i,val: %6.3f [%i %i]",c1,get.key[1],get.data[sv],mesh_proc(0,get.key[1]),mesh_proc(1,get.key[1]));
			get.data*=1/get.data.data[sv];														//divide over valence (later replace by one over table ?)
			typ.data[1]=MESH_ZIPIND[mesh_proc(0,get.key.data[1])];		//lookup to figure out to which vertex class (TOC,PRIS,CUBE) to add
			zip->find(typ,tmp); tmp+=get.data; zip->prior(typ,tmp);		//KUKUK 1
			typ.data[1]=MESH_ZIPIND[mesh_proc(1,get.key.data[1])];
			zip->find(typ,tmp); tmp-=get.data; zip->prior(typ,tmp);		//KUKUK 2 (the difference between 1 and 2 is only - + !!!)				
		}
	}
}

template <class type> void mesh<type>::sub_lin(void) {
	int c1,c2,c3,c4,c5,cnt=0,ofs,hard,size[MESH_MAX],sz[2];		
	mesh_hash_numel=zip->numel/2+1;
	adr=new hash<array<int>,int>(mesh_hash_numel,mesh_hash);
	array<int> *badge[5];												//duplicate vertices (carefully :-)
	for (c1=1;c1<5;c1++) {c2=c1*2; badge[c1]=new array<int>(1,&c2,0); }
	c2=2;
	array<int> typ(1,&c2,0);							//to get hash adr
	for (c1=0;c1<MESH_MAX;c1++) {								//loop over all shape types
		typ.data[1]=MESH_CLASS[c1];								//1=2nd entry gets face type
		ind[c1]->stretch(MESH_REPOSIT[c1],0);			//fill with zeros
		ofs=MESH_COMP[c1];												//init for current face type (zeros begin here)		
		for (c2=1;c2<5;c2++) {										//edges, quads, octa midpoint and cube midp.
			if (mesh_rep[cnt]) {				
				for (c3=0;c3<mesh_rep[cnt]->numel;c3+=2*c2) {					//all edges/quads/... 					
					hard=0;					
					for (c4=0;c4<ind[c1]->numel;c4+=ind[c1]->size[0]) {				//all of curr type (already stretched)
						for (c5=0;c5<2*c2;c5++)						//might be replaced by queue ???
							badge[c2]->data[c5]=ind[c1]->data[c4+mesh_rep[cnt]->data[c3+c5]];
						ind[c1]->data[ofs+hard]=typ.data[0]=getAdr(badge[c2]);  //get (new/old) address in zip hash, badge returns sorted (sometimes)
						mean(typ,badge[c2]); 																		//maybe insert mean vertex
						hard+=ind[c1]->size[0];
					}
					ofs++;															//one vertex was added per edges/quads/...
				}				
			}																				//if else => NULL
			cnt++;																	//now edge->quad->...
		}
	}//<-- here
	for (c1=1;c1<5;c1++) delete badge[c1];
	//assign new index arrays (just as many as needed) ... create nu faces, first init space, than marathon without hurldes
	for (c1=0;c1<MESH_MAX;c1++)
		size[c1]=ind[c1]->size[1]*MESH_REPRO[c1];
	size[MESH_TETR]+=ind[MESH_OCTA]->size[1]*8; //correction
	size[MESH_OCTA]+=ind[MESH_TETR]->size[1];   //size contains now right lengths
	array<int> *tind[MESH_MAX];
	for (c1=0;c1<MESH_MAX;c1++) {
		sz[0]=MESH_COMP[c1]; sz[1]=size[c1];
		tind[c1]=new array<int>(2,sz,0);		
		for (c2=0;c2<ind[c1]->numel;c2+=ind[c1]->size[0])
			for (c3=0;c3<mesh_uni[c1].numel;c3++)				
				*tind[c1]<=ind[c1]->data[c2+mesh_uni[c1].data[c3]];				
	}
	for (c2=0;c2<ind[MESH_OCTA]->numel;c2+=ind[MESH_OCTA]->size[0]) 							//for all old octas
		for (c3=0;c3<mesh_bas[MESH_OCTA].numel;c3++)
			*tind[MESH_TETR]<=ind[MESH_OCTA]->data[c2+mesh_bas[MESH_OCTA].data[c3]];	//add tetras			
	for (c2=0;c2<ind[MESH_TETR]->numel;c2+=ind[MESH_TETR]->size[0]) 							//for all old tetras
		for (c3=0;c3<mesh_bas[MESH_TETR].numel;c3++)
			*tind[MESH_OCTA]<=ind[MESH_TETR]->data[c2+mesh_bas[MESH_TETR].data[c3]];	//add octas	
	for (c1=0;c1<MESH_MAX;c1++) {
		delete ind[c1]; ind[c1]=tind[c1];					//swap buffers by pointer exchange
	}
	delete adr; 
	//printf("[%i/%i]",count,zip->table->numel);
}

template <class type> void mesh<type>::sub_avg(void) {	//this code might contain some vestiges
	int c1,c2,c3,c4,al,ml,tmp=count,sv=arr->size[0],svT=sizeof(type)*arr->size[0];
	array<char> dim(1,&tmp,3);											//0,1,2,3
	array<char> val(1,&tmp,0);
	array<type> wgt(1,&tmp,0);											//a little inefficient, to perform linear combination of final averaging
	array<int> wgi(1,&tmp,0);
	array<char> occ(1,&tmp,0);
	array<int>  typ(1,new int(2),0);								//zip hash key
	matrix<type> vrt(sv,1);
	for (c1=0;c1<MESH_MAX;c1++) {
		typ.data[1]=MESH_CLASS[c1];										//hash table is used as: [ vertex index | type of shape belonging to ]
		for (c2=0;c2<ind[c1]->numel;c2+=ind[c1]->size[0]) { // skip from 1 to end of curr ind						
			matrix<type> mat(sv,MESH_COMP[c1]);					//this has to stay in this loop!
			for (c3=0;c3<ind[c1]->size[0];c3++) {				//visit each vertex of curr face
				typ.data[0]=ind[c1]->data[c2+c3];
				zip->find(typ,vrt);
				memcpy(&mat.data[sv*c3],vrt.data,svT);
			}
			mat*=*mesh_avg[c1];													//the shape vertices get multiplied by THE (smoothing) matrix
			for (c3=0;c3<ind[c1]->size[0];c3++)					//for all vertices of shape (under alcohol, things need to be commented thoroughly)
				if (MESH_DIMS[c1]<=dim.data[al=ind[c1]->data[c2+c3]]) {		//if needed... (al !!! address of vertex in arr, note that dim is initially 3 !)
					dim.data[al]=MESH_DIMS[c1];
					val.data[al]++;													//...add to valence (only necessary for correction term, kinda redundant but prob faster)...
					wgt.data[al]+=(type)MESH_AVERAGEC[c1];					
					occ.data[al]|=MESH_OCCUP[c1];						//remember contrib. Possible combinations are MESH_QUAD_ or MESH_TRIA_|MESH_QUAD_ and so on
					ml=c3%mat.size[1];											//modulo trick for edge, quad and cube (matrix has constant entries)
					ml*=sv;					
					al*=sv;										
					for (c4=0;c4<sv;c4++)										//hard plug into array						
#ifndef MESH_FASTAVG						
						arr->data[al+c4]+=mat.data[ml+c4]*(type)MESH_AVERAGEC[c1];		//...to all coordinates (not efficient but saves space)
#else
						arr->data[al+c4]+=mat.data[ml+c4];		//...to all coordinates (not efficient but saves space)
#endif
				}
		}
	}
	al=0;																						//al is just a counter this time
	type f,w;																				//peace of code from here performs CC and Loop correction
	for (c1=0;c1<arr->numel;c1+=arr->size[0]) {
#ifndef MESH_FASTAVG				
		f=1/wgt.data[al];							//do the inverse of sum of weights used at vertex al(==)c1
#else
		f=1/(type)val.data[al];
#endif
		for (c2=0;c2<sv;c2++) arr->data[c1+c2]*=f;		//do averaging of topological identical vertices
		for (c3=MESH_TRIA;c3<MESH_TETR;c3++)
			if (occ.data[al]==MESH_OCCUP[c3]) {					//this time specialized
				typ.data[0]=al;
				typ.data[1]=MESH_CLASS[c3];
				zip->find(typ,vrt);												//"old" vertex position
				f=mesh_cor[c3]->data[val.data[al]];				//for new vertex pos
				w=1-f;																		//for old vertex pos
				for (c2=0;c2<sv;c2++) {
					arr->data[c1+c2]*=f;										//for new vertex pos
					arr->data[c1+c2]+=vrt.data[c2]*w;				//kalman filter order 0, no noise, adaptive least squares
				}
			}
		al++;
	}	
}
//--------------------------------------------------------------------------------------
template <class type> bool mesh<type>::match(array<int> &in1,array<int> &in2,array<int> &ord) {
	int lmn=MIN(in1.numel,in2.numel);
	ord=array<int>(1,&lmn,0);
	if (in1.numel==2) {											//=>edge vs edge
		if (in1[1]==in2[1]) {
			memcpy(ord.data,in1.data,8);			
			return true; } else return false;
	}
	if (in2.numel==3) {											//=>tri vs tri
		if (in1[1]==in2[2]&&in1[2]==in2[1]) {	//criterion is due to orientation
			memcpy(ord.data,in1.data,12);
			return true; } else return false;
	}
	if (in1.numel==3) {											//=>tri vs quad //assume no special alignment yet		
		int gia;
		if (in2.size[1]=MESH_PRIS) {					//if in2 comes from pris then detect type A/B. remove giant			
			in2.max(gia);												//by agreement in jsub_cpp.m in2 carries quad of pris			
			if (in2.data[gia]&MESH_GIANT) in2.data[gia]^=MESH_GIANT; else gia++; //beyond human thought
			gia%=2;															//easy compared to previous
		}
		lmn=5;
		array<int> stq(1,&lmn,0);
		memcpy(stq.data,&in1.data[1],8); memcpy(&stq.data[2],&in2.data[1],12);		
		stq.unique();
		if (stq.numel==3) {										//triangle is entirely in quad			
			int c1=0,c2,pos;			
			while (in1.hasAt(in2.data[c1],pos)) c1++;	//at the end c1 indexes element of quad not in tri.
			pos=c1;
			for (c2=0;c2<3;c2++) ord[c2]=in2[++c1%4];	//used to loop downward								
			if (in2.size[1]=MESH_PRIS) {				
				pos+=gia;
				if (pos%2) {lmn=ord[0]; ord[0]=ord[2]; ord[2]=lmn; }
			}
			return true; } else return false;
	}
	if (in1.numel==4) {											//=>quad vs quad (if all fail this ought be true, so if not really necessary)
		in1.min(lmn);													//by agreement in jsub_cpp.m in2 carries quad of pris
		if (lmn) return false;								//quad is duplicate from TRUE quad
		lmn=6;
		array<int> stq(1,&lmn,0);
		memcpy(stq.data,&in1.data[1],12); memcpy(&stq.data[3],&in2.data[1],12);
		stq.unique();
		if (stq.numel<5) {										//quad is oon quad, pris might carry giant
			if (in2[2]&MESH_GIANT) {
				memcpy(ord.data,in2.data,16);
				ord[2]=in1[2];										//get rid of giant!
			} else 
				memcpy(ord.data,in1.data,16);			//out little matrix is commutative to a certain extent
			return true; } else return false;	
	}
	return false;
}

template <class type> int mesh<type>::numVertEst(void) {
	int c1,est=arr->size[1];
	for (c1=0;c1<MESH_MAX;c1++)
		est+=ind[c1]->size[1]*MESH_ADD_VPF[c1];
	est+=arr->size[1]+100;
	return est;
}

template <class type> int mesh<type>::getAdr(array<int> *badge) {
	int pos=count;
	if (badge->numel!=4) {
		badge->sort();
		if (adr->store(*badge,pos)) count++;
	} else {
		int c1=2;
		array<int> ed1(1,&c1,0),ed2(1,&c1,0);
		ed1.data[0]=badge->data[0]; ed1.data[1]=badge->data[2]; ed1.sort();
		ed2.data[0]=badge->data[1]; ed2.data[1]=badge->data[3]; ed2.sort();
		if (!adr->find(ed1,pos)&&!adr->find(ed2,pos)) {
			pos=count++;
			adr->store(ed1,pos);
			adr->store(ed2,pos);
		}
	}
	return pos;
}

template <class type> array<int> mesh<type>::boundary(array<int> &mat) {
	int c1;
	array<int> key(1,&mat.numel,0);
	key.data[0]=mat.data[0];
	for (c1=1;c1<mat.numel-1;c1++) key.data[c1]=mat.data[mat.numel-c1-1];
	key.data[c1]=mat.data[c1];
	return key;
}

template <class type> void mesh<type>::mean(array<int> &mat,array<int> *badge) {
	matrix<type> tmp(arr->size[0],1);		//temporary vertex
	if (!zip->find(mat,tmp)) {					//mat has [adress of vertex | TETR or ...]
		int c1,pos=mat.data[0];						//mean vertex insertion point
		matrix<type> res(arr->size[0],1); //
		for (c1=0;c1<badge->numel;c1++) { //accesses all old vertices (that badge points to)
			mat.data[0]=badge->data[c1];		//injects vertex adress
			zip->find(mat,res);							//should be true since access of an old face!!
			tmp+=res;												//add ....
		}
		tmp*=1/(type)badge->numel;				//now vertices have accumulate so we can mean it
		mat.data[0]=pos;									//mat is new vertex adress again (was that originally)		
		zip->store(mat,tmp);							//store (of course)
	}
}
//--------------------------------------------------------------------------------------
//if the following operator causes trouble, just comment it out
template <class type> ostream& operator<< (ostream& os, mesh<type>& a);
template <class type> ostream& operator<< (ostream& os, mesh<type>& a) {
	int c1;
	for (c1=0;c1<MESH_MAX;c1++) os << *a.ind[c1];
	if (a.arr) os << *a.arr; else os << "vertex array empty!" << endl;
	//if (a.zip) os << *a.zip;
	return os;
}

#endif